/* Copyright (C) Louisiana State University (2014-2017)
 *               Cardiff University (2017-2022)
 *
 * This file is part of GWpy.
 *
 * GWpy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GWpy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This file contains sphinx-specific extensions to the core GWpy CSS
 */

/* change font for parameter names for callables */
.field-list p strong {
		font-family: "Roboto Mono", "Courier New", Courier, monospace;
}

/* remove left border from parameter descriptions for callables
 * and indent more */
[dir=ltr] .md-typeset blockquote {
		border-left: 0;
		padding-left: 1em;
}

/* remove padding for code blocks forced in by sphinx-panels */
.literal-block-wrapper.container {
		padding-left: 0;
		padding-right: 0;
}
